# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/01_test.ipynb (unless otherwise specified).

__all__ = ['test_erc', 'test_drc', 'validate_bom', 'test_notebooks']

# Cell
import os
import subprocess

from fastcore.script import *
from nbdev.test import *
from nbdev.test import nbglob, num_cpus, parallel, _test_one, Path
import pandas as pd
import pandera as pa
from kicad_helpers import *
from .utilities import _set_root, _print_cmd_output

# Cell
def test_erc(root="."):
    returncode = 0
    config = f".kicad_helpers_config/erc.yaml"
    try:
        output = run_kibot_docker(config=config, root=root)
    except subprocess.CalledProcessError as e:
        returncode = e.returncode
        print(e.output.decode("utf-8"))
        print(f"returncode = { returncode }")
    erc_path = os.path.join(root, get_project_name(root) + ".erc")
    if os.path.exists(erc_path):
        os.remove(erc_path)
    assert returncode == 0

# Cell
def test_drc(root="."):
    returncode = 0
    config = f".kicad_helpers_config/drc.yaml"
    try:
        output = run_kibot_docker(config=config, root=root)
    except subprocess.CalledProcessError as e:
        returncode = e.returncode
        print(e.output.decode("utf-8"))
        print(f"returncode = { returncode }")
    drc_path = os.path.join(root, "drc_result.rpt")
    if os.path.exists(drc_path):
        os.remove(drc_path)
    assert returncode == 0

# Cell
def validate_bom(root="."):
    df = pd.read_csv(get_bom_path(root))
    schema = pa.DataFrameSchema({
        "Refs": pa.Column(str),
        "Quantity": pa.Column(int),
        "MPN": pa.Column(str),
        "Manufacturer": pa.Column(str),
        "datasheet": pa.Column(str, nullable=True, coerce=True),
        "footprint": pa.Column(str),
        "value": pa.Column(str),
    })

    return schema.validate(df)

# Cell
@call_parse
def test_notebooks(fname:Param("A notebook name or glob to convert", str)=None,
                   flags:Param("Space separated list of flags", str)=None,
                   n_workers:Param("Number of workers to use", int)=None,
                   verbose:Param("Print errors along the way", bool_arg)=True,
                   timing:Param("Timing each notebook to see the ones are slow", bool)=False,
                   pause:Param("Pause time (in secs) between notebooks to avoid race conditions", float)=0.5,
                   root:Param("project root directory", str)="."):
    """Test all notebooks matching `fname` in parallel, passing along `flags`"""
    root = _set_root(root)
    if flags is not None: flags = flags.split(' ')
    if fname is None:
        fname = os.path.join(root, "tests", "*.ipynb")
    files = nbglob(fname, recursive=False)
    files = [Path(f).absolute() for f in sorted(files)]
    assert len(files) > 0, "No files to test found."
    if n_workers is None: n_workers = 0 if len(files)==1 else min(num_cpus(), 8)
    # make sure we are inside the tests folder
    os.chdir(os.path.join(root, "tests"))
    results = parallel(_test_one, files, flags=flags, verbose=verbose, n_workers=n_workers, pause=pause)
    passed,times = [r[0] for r in results],[r[1] for r in results]
    if all(passed): print("All tests are passing!")
    else:
        msg = "The following notebooks failed:\n"
        raise Exception(msg + '\n'.join([f.name for p,f in zip(passed,files) if not p]))
    if timing:
        for i,t in sorted(enumerate(times), key=lambda o:o[1], reverse=True):
            print(f"Notebook {files[i].name} took {int(t)} seconds")


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"update_templates": "00_actions.ipynb",
         "add_badges": "00_actions.ipynb",
         "install_git_filters": "00_actions.ipynb",
         "update_gitignore": "00_actions.ipynb",
         "update_project": "00_actions.ipynb",
         "sch_to_bom": "00_actions.ipynb",
         "bom_to_sch": "00_actions.ipynb",
         "export_manufacturing": "00_actions.ipynb",
         "export_sch": "00_actions.ipynb",
         "export_pcb": "00_actions.ipynb",
         "run_erc": "00_actions.ipynb",
         "run_drc": "00_actions.ipynb",
         "set_date": "00_actions.ipynb",
         "set_revision": "00_actions.ipynb",
         "test_erc": "01_test.ipynb",
         "test_drc": "01_test.ipynb",
         "validate_bom": "01_test.ipynb",
         "test_notebooks": "01_test.ipynb",
         "get_git_root": "02_utilities.ipynb",
         "setup_test_repo": "02_utilities.ipynb",
         "get_project_name": "02_utilities.ipynb",
         "get_project_metadata": "02_utilities.ipynb",
         "get_schematic_path": "02_utilities.ipynb",
         "get_bom_path": "02_utilities.ipynb",
         "get_board_path": "02_utilities.ipynb",
         "get_manufacturers": "02_utilities.ipynb",
         "get_gitignore_list": "02_utilities.ipynb",
         "in_gitignore": "02_utilities.ipynb",
         "run_docker_cmd": "02_utilities.ipynb",
         "run_kibot_docker": "02_utilities.ipynb",
         "get_board_metadata": "02_utilities.ipynb",
         "update_board_metadata": "02_utilities.ipynb",
         "get_schematic_metadata": "02_utilities.ipynb",
         "update_schematic_metadata": "02_utilities.ipynb",
         "github_badge": "02_utilities.ipynb",
         "kitspace_badge": "02_utilities.ipynb"}

modules = ["actions.py",
           "test.py",
           "utilities.py"]

doc_url = "https://ryanfobel.github.io/kicad-helpers/"

git_url = "https://github.com/ryanfobel/kicad-helpers/tree/main/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/00_actions.ipynb (unless otherwise specified).

__all__ = ['update_templates', 'add_badges', 'install_git_filters', 'update_gitignore', 'update_project', 'sch_to_bom',
           'bom_to_sch', 'export_manufacturing', 'export_sch', 'export_pcb', 'run_erc', 'run_drc', 'set_date',
           'set_revision']

# Cell
import os
import sys
import pkg_resources
import subprocess
from pprint import pprint
import datetime as dt

import jinja2
from fastcore.script import *
import pandas as pd
from kifield.kifield import explode

from kicad_helpers import *
from .utilities import _set_root, _run_cmd, _print_cmd_output

# Cell
def update_templates(v:Param("verbose", bool),
                     overwrite:Param("overwrite existing templates", bool),
                     root:Param("project root directory", str)="."):
    """Install templates from the `kicad_helpers/templates` directory (ignoring
    anything in the project's `.gitignore` list).
    """
    templates_path = os.path.abspath(pkg_resources.resource_filename('kicad_helpers', 'templates'))
    root = _set_root(root)
    metadata = get_project_metadata(root)
    file_list = []
    exists_flag = False
    for root_, dirs, files in os.walk(templates_path):
        if len(files):
            for file in files:
                path = os.path.join(root_[len(templates_path) + 1:], file)
                if not in_gitignore(path, root):
                    src_path = os.path.abspath(os.path.join(templates_path, path))
                    dst_path = os.path.abspath(os.path.join(root, path))

                    # Create the `dst_path` directory if it doesn't exist
                    os.makedirs(os.path.split(dst_path)[0], exist_ok=True)

                    if os.path.exists(dst_path):
                        if not overwrite:
                            print(f"{ path } already exists")
                            exists_flag = True
                            continue

                    with open(src_path) as f:
                        template = jinja2.Template(f.read())

                    with open(dst_path, "w") as f:
                        if v:
                            print(f"Render { path } template.")
                        f.write(template.render(**metadata))

    if not overwrite and exists_flag:
        print("To overwrite existing files, use the --overwrite flag.")

# Cell
def _add_lines_to_file(root, file_name, lines, v=False, prepend=False):
    file_path = os.path.join(root, file_name)
    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            file_lines = f.readlines()
    else:
        file_lines = []

    for line in lines:
        if line not in file_lines:
            if v:
                print(f"Append \"{ line.strip() }\" to { file_path }")
            if prepend:
                file_lines.insert(0, line)
            else:
                file_lines.append(line)
        elif v:
            print(f"\"{ line.strip() }\" already exists in { file_path }")

    with open(file_path, "w") as f:
        f.writelines(file_lines)

# Cell
@call_parse
def add_badges(root:Param("project root directory", str)=".",
               v:Param("verbose", bool)=False,
               github:Param("github", bool)=False,
               kitspace:Param("kitspace", bool)=False):
    """Add badges to the README.md file."""
    root = _set_root(root)
    badges = []
    if kitspace:
        badges.append(kitspace_badge(root))
    if github:
        badges.append(github_badge(root))
    _add_lines_to_file(root,
                       "README.md",
                       [line + "\n" for line in badges],
                       v,
                       prepend=True
    )

# Cell
_gitattributes_list = ["*.pro filter=kicad_project",
    "*.sch filter=kicad_sch"
]

def install_git_filters(root=".", v=False):
    """Install git filters to prevent insignificant changes to the kicad
    `*.pro` and `*.sch` files from being tracked by git.

    See: https://jnavila.github.io/plotkicadsch/
    """
    root = _set_root(root)

    # Add filters for kicad project and schematic files
    _add_lines_to_file(root,
                       ".gitattributes",
                       [line + "\n" for line in _gitattributes_list],
                       v
    )
    if v:
        print("Add filters to git config.")

    # Add filters to the project's git config
    _run_cmd(f"cd { root } && git config filter.kicad_project.clean \"sed -E 's/^update=.*$/update=Date/'\"")
    _run_cmd(f"cd { root } && git config filter.kicad_project.smudge cat")

    # This seems to break ERC checking. Disable for now...
    #_run_cmd(f"cd { root } && git config filter.kicad_sch.clean \"sed -E 's/#(PWR|FLG)[0-9]+/#\1?/'\"")
    #_run_cmd(f"cd { root } && git config filter.kicad_sch.smudge cat")

# Cell

_gitignore_list = ["_autosave*",
    "*bak",
    "*.xml",
    ".ipynb_checkpoints",
    "*-erc.txt",
    "*-drc.txt",
    "kibot_errors.filter"
]

def update_gitignore(root=".", v=False):
    """Add the following entries to the `.gitignore` file:
    ```
        _autosave*
        *bak
        *.xml
        .ipynb_checkpoints
        *-erc.txt
        *-drc.txt
        kibot_errors.filter
    ```
    """
    # Add filters for kicad project and schematic files
    _add_lines_to_file(root,
                       ".gitignore",
                       [line + "\n" for line in _gitignore_list],
                       v
    )

# Cell
@call_parse
def update_project(v:Param("verbose", bool),
                   overwrite:Param("overwrite existing templates", bool),
                   root:Param("project root directory", str)="."):
    """Setup a new project (or update an existing project) with templates from
    the `kicad_helpers/templates` directory.
    Also installs git filters to prevent insignificant changes to the kicad
    `*.pro` and `*.sch` files from being tracked by git (see
    https://jnavila.github.io/plotkicadsch/ for more details).
    """
    update_templates(v=v, overwrite=overwrite, root=root)
    install_git_filters(root=root, v=v)
    update_gitignore(root=root, v=v)

# Cell
@call_parse
def sch_to_bom(root:Param("project root directory", str)=".",
               v:Param("verbose", bool)=False,
               overwrite:Param("update existing schematic", bool)=False):
    """Update/create BOM from KiCad schematic.
    """
    root = _set_root(root)
    cmd = f"{ sys.executable } -m kifield -r --nobackup --overwrite --group -x { get_schematic_path(root) } -i { get_bom_path(root) }"
    if v:
        print(cmd)
    _print_cmd_output(cmd)

    # Open the BOM
    df = pd.read_csv(get_bom_path(root))

    # Calculate the quantity of each part
    df["Quantity"] = [len(explode(refs)) for refs in df["Refs"]]

    # Make quantity the second column
    if df.columns[-1] == "Quantity":
        df = df[[df.columns[0]] + ["Quantity"] + list(df.columns[1:-1])]

    # Sort alphabetically by Ref
    df.sort_values(by="Refs", inplace=True)
    df.to_csv(get_bom_path(root), index=False)

# Cell
@call_parse
def bom_to_sch(root:Param("project root directory", str)=".",
               v:Param("verbose", bool)=False,
               overwrite:Param("update existing schematic", bool)=False):
    """Update KiCad schematic from BOM file.
    """
    root = _set_root(root)
    cmd = f"{ sys.executable } -m kifield -r --nobackup --overwrite --fields ~Quantity -x { get_bom_path(root) } -i { get_schematic_path(root) }"
    if v:
        print(cmd)
    _print_cmd_output(cmd)

# Cell
@call_parse
def export_manufacturing(root:Param("project root directory", str)=".",
                         manufacturer:Param(f"\"default\" or manufacturer name", str)="default",
                         v:Param("verbose", bool)=False,
                         output:Param("output path relative to ROOT")="."):
    """Export manufacturing files (gerber, drill, and position) by running
    KiBot in a local docker container.
    """
    root = _set_root(root)
    if manufacturer not in get_manufacturers(root):
        raise RuntimeError(f"MANUFACTURER must be one of the following: { ', '.join(get_manufacturers(root)) }.")

    config = f".kicad_helpers_config/manufacturers/{ manufacturer }.yaml"
    run_kibot_docker(config=config, root=root, v=v, output=output)

# Cell
@call_parse
def export_sch(root:Param("project root directory", str)=".",
               ext:Param(f"svg or pdf", str)="pdf",
               v:Param("verbose", bool)=False,
               output:Param("output path relative to ROOT")="."):
    """Export the schematic by running KiBot in a local docker container.
    """
    root = _set_root(root)
    supported_types = ["svg", "pdf"]

    if ext not in supported_types:
        raise RuntimeError(f"EXT must be one of: { ','.join(supported_types) }.")

    config = f".kicad_helpers_config/sch_{ ext }.yaml"
    run_kibot_docker(config=config, root=root, v=v, output=output)

# Cell
@call_parse
def export_pcb(root:Param("project root directory", str)=".",
               ext:Param(f"svg or pdf", str)="pdf",
               v:Param("verbose", bool)=False,
               output:Param("output path relative to ROOT")="."):
    """Export the pcb layout by running KiBot in a local docker container.
    """
    root = _set_root(root)
    supported_types = ["svg", "pdf"]

    if ext not in supported_types:
        raise RuntimeError(f"EXT must be one of: { ','.join(supported_types) }.")

    config = f".kicad_helpers_config/pcb_{ ext }.yaml"
    run_kibot_docker(config=config, root=root, v=v, output=output)

# Cell
@call_parse
def run_erc(root:Param("project root directory", str)=".",
            v:Param("verbose", bool)=False):
    """Run electrical rules check (ERC) to verify schematic connections. It
    checks for output pin conflicts, missing drivers and unconnected pins.
    Print the report to `stdout`.
    """
    root = _set_root(root)
    cmd = f"eeschema_do run_erc { get_schematic_path(root)[len(root) + 1:] } ."
    try:
        output = run_docker_cmd(cmd,
           workdir=os.path.abspath(root),
           container="setsoft/kicad_auto_test:latest",
           v=v
        )
    except subprocess.CalledProcessError as e:
        output = e.output.decode("utf-8")
    if v:
        print(output)
    erc_path = os.path.join(root, get_project_name(root) + ".erc")
    with open(erc_path, "r") as f:
        erc = f.read()
    os.remove(erc_path)
    print(erc)

# Cell
@call_parse
def run_drc(root:Param("project root directory", str)=".",
            v:Param("verbose", bool)=False):
    """Run design rules check (DRC) and print the report to `stdout`.
    """
    root = _set_root(root)
    cmd = f"pcbnew_do run_drc { get_board_path(root)[len(root) + 1:] } ."
    try:
        output = run_docker_cmd(cmd,
            workdir=os.path.abspath(root),
            container="setsoft/kicad_auto_test:latest",
            v=v
        )
    except subprocess.CalledProcessError as e:
        output = e.output.decode("utf-8")
    if v:
        print(output)
    drc_path = os.path.join(root, "drc_result.rpt")
    with open(drc_path, "r") as f:
        drc = f.read()
    os.remove(drc_path)
    print(drc)

# Cell
@call_parse
def set_date(date:Param("date (defaults to today's date)", str)=None,
             root:Param("project root directory", str)=".",
             v:Param("verbose", bool)=False):
    """Set the date in all schematic and board files.
    """
    root = _set_root(root)
    if date is None:
        date = dt.datetime.now().date().isoformat()
    update_schematic_metadata({"Date": date},
                              root=root, all_sheets=True)
    update_board_metadata({"date": date}, root=root)

# Cell
@call_parse
def set_revision(revision:Param("revision", str),
                 root:Param("project root directory", str)=".",
                 v:Param("verbose", bool)=False):
    """Set the revision in all schematic and board files.
    """
    root = _set_root(root)
    update_schematic_metadata({"Rev": revision},
                              root=root, all_sheets=True)
    update_board_metadata({"rev": revision}, root=root)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/02_utilities.ipynb (unless otherwise specified).

__all__ = ['get_git_root', 'setup_test_repo', 'get_project_name', 'get_project_metadata', 'get_schematic_path',
           'get_bom_path', 'get_board_path', 'get_manufacturers', 'get_gitignore_list', 'in_gitignore',
           'run_docker_cmd', 'run_kibot_docker', 'get_board_metadata', 'update_board_metadata',
           'get_schematic_metadata', 'update_schematic_metadata', 'github_badge', 'kitspace_badge']

# Cell
import glob
import os
import re
import subprocess
import tempfile
from pprint import pprint
import urllib.parse

import git
from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
from fastcore.script import *

# Cell
def _run_cmd(cmd):
    return subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True).decode("utf-8")

def _print_cmd_output(cmd):
    print(_run_cmd(cmd))

# Cell
def get_git_root(path="."):
    # Find the current projects' root directory
    git_repo = git.Repo(path, search_parent_directories=True)
    return git_repo.git.rev_parse("--show-toplevel").replace("/", os.path.sep)

# Cell
@call_parse
def setup_test_repo(root:Param("project root directory", str)="_temp"):
    """Setup a test KiCad repository to test against.
    """
    if not os.path.exists(root):
        _run_cmd(f"git clone --recursive https://github.com/sci-bots/dropbot-40-channel-HV-switching-board.kicad { root }")

# Cell
def _set_root(root):
    """If `root` is the default value ("."), use the project's git root
    or override with the environment variable `KH_PROJECT_ROOT` if it
    exists.
    """
    if root == ".":
        # Use defaults
        root = get_git_root(".")

        # Override with environment variable if set
        root = os.getenv("KH_PROJECT_ROOT", root)
    return root

# Cell
def get_project_name(root="."):
    """Get the project name based on the name of the KiCad `*.pro` file.
    """
    root = _set_root(root)
    return os.path.splitext(os.path.split(glob.glob(os.path.join(root, "*.pro"))[0])[1])[0]

# Cell
def get_project_metadata(root="."):
    """Get the project metatdata from the `kitspace.yaml` file.
    """
    root = _set_root(root)

    # Default metadata if there's no existing `kitspace.yaml` file.
    metadata = {"summary": "A description for your project",
                "site": "https://example.com # A site you would like to link to (include http:// or https://)",
                "color": "black"
    }

    try:
        # If there's an existing `kicad.yaml` file, those settings override the defaults.
        with open(os.path.join(root, "kitspace.yaml")) as f:
            metadata.update({k: v for k, v in load(f, Loader=Loader).items() if k in ["summary", "site", "color"]})
    except FileNotFoundError:
        pass

    # Add the project name
    metadata["project_name"]=get_project_name(root)
    return metadata


# Cell
def get_schematic_path(root="."):
    """Get the path to the KiCad schematic.
    """
    root = _set_root(root)
    return os.path.join(root, get_project_name(root) + ".sch")

# Cell
def get_bom_path(root="."):
    """Get the path to the BOM.
    """
    root = _set_root(root)
    return os.path.join(root, "manufacturing", "default", get_project_name(root) + "-BOM.csv")

# Cell
def get_board_path(root="."):
    """Get the path to the KiCad board file.
    """
    root = _set_root(root)
    return os.path.join(root, get_project_name(root) + ".kicad_pcb")

# Cell
def get_manufacturers(root="."):
    """Get the supported manufacturers.
    """
    root = _set_root(root)
    return [os.path.split(f)[1][:-5] for f in glob.glob(os.path.join(root, ".kicad_helpers_config", "manufacturers", "*.yaml"))]

# Cell
def get_gitignore_list(root="."):
    root =_set_root(root)
    with open(f"{ root }/.gitignore") as f:
        gitignore = f.readlines()
    return [line.strip() for line in gitignore]

# Cell
def in_gitignore(filename, root="."):
    root = _set_root(root)
    try:
        cmd = f"cd { root } && git check-ignore --no-index { filename }"
        if len(_run_cmd(cmd)):
            return True
    except subprocess.CalledProcessError:
        pass
    return False

# Cell
def run_docker_cmd(cmd,
                   workdir,
                   container,
                   v=False):
    """
    Run a command in a docker container under a UID mapped to the current user.
    This ensures that the current user is owner of any files created in the
    workdir.
    """
    UID = subprocess.check_output("id -u", shell=True).decode("utf-8").strip()
    docker_cmd = (f"docker run --rm -v { workdir }:/workdir --workdir=\"/workdir\" "
        f"{ container } "
        f"/bin/bash -c \"useradd --shell /bin/bash -u { UID } -o -c '' -m docker && "
        f"runuser docker -c '{ cmd }'\""
    )
    if v:
        print(docker_cmd)
    return subprocess.check_output(docker_cmd, stderr=subprocess.STDOUT, shell=True)

# Cell
def run_kibot_docker(config:Param(f"KiBot configuation file", str),
                     root:Param("project root directory", str)=".",
                     v:Param("verbose", bool)=False,
                     output:Param("output path relative to ROOT")="."):
    """
    Run KiBot in a local docker container.
    """
    root = _set_root(root)
    if os.path.abspath(output) == output:
        raise RuntimeError(f"OUTPUT cannot be an absolute path; it must be relative to ROOT={ root }.")

    cmd = (f"kibot -c { config } "
       f"-e { get_schematic_path(root)[len(root) + 1:] } "
       f"-b { get_board_path(root)[len(root) + 1:] } "
       f"-d { output }"
    )
    run_docker_cmd(cmd,
                   workdir=os.path.abspath(root),
                   container="setsoft/kicad_auto_test:latest",
                   v=v
    )

# Cell
def get_board_metadata(root="."):
    """Get metadata from the `*.kicad_pcb` board file.
    """
    root = _set_root(root)
    with open(get_board_path(root), 'r') as f:
        board = f.read()
    matches = re.search("\(title_block.*title (?P<title>[^\)]*)\)"
        ".*date (?P<date>[^\)]*)\)"
        ".*rev (?P<rev>[^\)]*)\)"
        ".*company (?P<company>[^\)]*)\)",
        board, re.DOTALL
    )
    return matches.groupdict()

# Cell
def update_board_metadata(update_dict, root="."):
    """Update metadata in the `*.kicad_pcb` board file.
    """
    root = _set_root(root)
    with open(get_board_path(root), 'r') as f:
        board = f.read()

    for key, value in update_dict.items():
        if " " in value and not value.startswith('\"') and not value.endswith('\"'):
            value = '\"' + value + '\"'
        board, n = re.subn(f"(?s)(?P<pre>\(title_block.*{ key } )[^\)]*\)",
            f"\g<pre>{ value })",
            board
        )
        assert n == 1
    with open(get_board_path(root), 'w') as f:
        f.write(board)

# Cell
def get_schematic_metadata(root, filename=None):
    """Get metadata from a `*.sch` schematic file.
    """
    if filename is None:
        filename = get_schematic_path(root)
    elif os.path.abspath(filename) != filename:
        filename = os.path.join(root, filename)

    with open(filename, 'r') as f:
        schematic = f.read()

    matches = re.search("Title (?P<Title>[^\n]*)\n"
        ".*Date (?P<Date>[^\n]*)\n"
        ".*Rev (?P<Rev>[^\n]*)\n"
        ".*Comp (?P<Comp>[^\n]*)\n",
        schematic, re.DOTALL
    )

    return matches.groupdict()

# Cell
def update_schematic_metadata(update_dict:dict,      # keys/values to update
                              root:str=".",          # project root directory
                              all_sheets:bool=True): # update subsheets
    """Update metadata in a `*.sch` schematic file.
    """
    root = _set_root(root)

    if all_sheets:
        files = glob.glob(os.path.join(root, "*.sch"))
    else:
        files = [get_schematic_path(root)]

    for file in files:
        with open(file, 'r') as f:
            schematic = f.read()

        for key, value in update_dict.items():
            if not value.startswith('\"') and not value.endswith('\"'):
                value = '\"' + value + '\"'
            schematic, n = re.subn(f"(?s)(?P<pre>{ key } )([^\n]*)\n",
                f"\g<pre>{ value }\n",
                schematic
            )
            assert n == 1
        with open(file, 'w') as f:
            f.write(schematic)

# Cell
def github_badge(root="."):
    root = _set_root(root)
    github_build_action = get_project_metadata(root)["site"] + "/actions/workflows/build.yml"
    return f"[![tests]({ github_build_action }{ '/badge.svg' })]({ github_build_action })"

# Cell
def kitspace_badge(root="."):
    root = _set_root(root)
    kitspace_link = ("https://kitspace.org/boards/" +
        get_project_metadata(root)["site"].replace("https://", "")
    )
    return ("[![kitspace](https://img.shields.io/website?down_color=red&"
        "down_message=no&label=kitspace&up_color=sucess&up_message=ok&"
        f"url={ urllib.parse.quote_plus(kitspace_link) })]("
        f"{ kitspace_link })"
    )